import unittest

from bussilab.wham import wham
from bussilab.coretools import TestCase

class TestWham(TestCase):
    def test_wham(self):
        import numpy as np
        a = np.exp(wham(np.array([[1, 10, 7], [2, 9, 6], [3, 8, 5]])).logW)
        self.assertAlmostEqual(np.sum((a-np.array([0.41728571, 0.39684866, 0.18586563]))**2), 0.0)
        b = np.exp(wham(np.array([[1, 10], [2, 9], [3, 8]]), traj_weight=(0.1, 0.2)).logW)
        self.assertAlmostEqual(np.sum((b-np.array([0.41728571, 0.39684866, 0.18586563]))**2), 0.0)
        c = np.exp(wham(np.array([[3, 5],
                                  [4, 4],
                                  [4, 4],
                                  [4, 4],
                                  [4, 4],
                                  [4, 4],
                                  [4, 4],
                                  [4, 4],
                                  [4, 4],
                                  [4, 4],
                                  [4, 4]])).logW)
        self.assertAlmostEqual(np.sum((c-np.array([
            0.06421006, 0.09357899, 0.09357899, 0.09357899, 0.09357899,
            0.09357899, 0.09357899, 0.09357899, 0.09357899, 0.09357899, 0.09357899
        ]))**2), 0.0)
        d = np.exp(wham(np.array([[3, 5], [4, 4]]), frame_weight=(1, 10)).logW)
        self.assertAlmostEqual(np.sum((d-np.array([[0.06421006, 0.93578994]]))**2), 0.0)
        e = np.exp(wham(np.array([[1, 5, 14], [2, 4.5, 12], [3, 4, 10]]),T=[1.0,0.5,2.0]).logW)
        self.assertAlmostEqual(np.sum((e-np.array([0.41728571, 0.39684866, 0.18586563]))**2), 0.0)

    def test_wham1s(self):
        import numpy as np
        # check the substituting version
        a = np.exp(wham(np.array([[1, 10, 7], [2, 9, 6], [3, 8, 5]]),method="substitute").logW)
        self.assertAlmostEqual(np.sum((a-np.array([0.41728571, 0.39684866, 0.18586563]))**2), 0.0)
        b = np.exp(wham(np.array([[1, 10], [2, 9], [3, 8]]), traj_weight=(1, 2),method="substitute").logW)
        self.assertAlmostEqual(np.sum((b-np.array([0.41728571, 0.39684866, 0.18586563]))**2), 0.0)
        c = np.exp(wham(np.array([[3, 5],
                                  [4, 4],
                                  [4, 4],
                                  [4, 4],
                                  [4, 4],
                                  [4, 4],
                                  [4, 4],
                                  [4, 4],
                                  [4, 4],
                                  [4, 4],
                                  [4, 4]])).logW)
        self.assertAlmostEqual(np.sum((c-np.array([
            0.06421006, 0.09357899, 0.09357899, 0.09357899, 0.09357899,
            0.09357899, 0.09357899, 0.09357899, 0.09357899, 0.09357899, 0.09357899
        ]))**2), 0.0)
        d = np.exp(wham(np.array([[3, 5], [4, 4]]), frame_weight=(1, 10),method="substitute").logW)
        self.assertAlmostEqual(np.sum((d-np.array([[0.06421006, 0.93578994]]))**2), 0.0)
        e = np.exp(wham(np.array([[1, 5, 14], [2, 4.5, 12], [3, 4, 10]]),T=[1.0,0.5,2.0],method="substitute").logW)
        self.assertAlmostEqual(np.sum((e-np.array([0.41728571, 0.39684866, 0.18586563]))**2), 0.0)

    def test_wham2(self):
        # test passing lists instead of np arrays
        import numpy as np
        a = np.exp(wham([[1, 10, 7], [2, 9, 6], [3, 8, 5]]).logW)
        self.assertAlmostEqual(np.sum((a-np.array([0.41728571, 0.39684866, 0.18586563]))**2), 0.0)
        b = np.exp(wham([[1, 10], [2, 9], [3, 8]], traj_weight=(1, 2)).logW)
        self.assertAlmostEqual(np.sum((b-np.array([0.41728571, 0.39684866, 0.18586563]))**2), 0.0)
        c = np.exp(wham([[3, 5], 
                         [4, 4],
                         [4, 4],
                         [4, 4],
                         [4, 4],
                         [4, 4],
                         [4, 4],
                         [4, 4],
                         [4, 4],
                         [4, 4],
                         [4, 4]]).logW)
        self.assertAlmostEqual(np.sum((c-np.array([
            0.06421006, 0.09357899, 0.09357899, 0.09357899, 0.09357899,
            0.09357899, 0.09357899, 0.09357899, 0.09357899, 0.09357899, 0.09357899
        ]))**2), 0.0)
        d = np.exp(wham(np.array([[3, 5], [4, 4]]), frame_weight=(1, 10)).logW)
        self.assertAlmostEqual(np.sum((d-np.array([[0.06421006, 0.93578994]]))**2), 0.0)

    def test_wham3(self):
        from bussilab import cli
        from bussilab.coretools import cd
        import os
        with cd(os.path.dirname(os.path.abspath(__file__))):
            try:
                os.remove("wham1.out")
            except FileNotFoundError:
                pass
            cli.cli("wham -b wham1.in -o wham1.out")
            self.assertEqualFile("wham1.out")
            os.remove("wham1.out")

    def test_wham4(self):
        import numpy as np
        # check if the algorithm is numerically robust if one of the biases if heavily offset
        for diffZ in (0.0,1e1,1e2,1e3,1e4,1e5,1e6):
            w=wham(2.0*np.array([[1,1-diffZ],[2,2-diffZ]]),T=2.0)
            self.assertAlmostEqual(w.logZ[1]-w.logZ[0],diffZ)
            self.assertAlmostEqual(w.logW[1]-w.logW[0],1.0)

    def test_wham4s(self):
        import numpy as np
        # check if the algorithm is numerically robust if one of the biases if heavily offset, with substitution
        for diffZ in (0.0,1e1,1e2,1e3,1e4,1e5,1e6):
            w=wham(2.0*np.array([[1,1-diffZ],[2,2-diffZ]]),method="substitute",T=2.0)
            self.assertAlmostEqual(w.logZ[1]-w.logZ[0],diffZ)
            self.assertAlmostEqual(w.logW[1]-w.logW[0],1.0)

    def test_wham5(self):
        import numpy as np
        # check if the algorithm is numerically robust if one frame has weight much smaller than another
        for diffW in (0.0,1e1,1e2,1e3,1e4,1e5,1e6):
            w=wham(2.0*np.array([[0.0,0.0],[diffW,diffW]]),T=2.0)
            self.assertAlmostEqual(w.logW[1]-w.logW[0],diffW)
            self.assertAlmostEqual(w.logZ[1]-w.logZ[0],0.0)

    def test_wham_restart(self):
        import numpy as np
        w0=wham(((0,10),(-4,11)),T=0.2)
        self.assertGreaterEqual(w0.nfev,8)
        w1=wham(((0,10),(-4,11)),T=0.2,logW=w0.logW)
        self.assertLessEqual(w1.nfev,2)
        w2=wham(((0,10),(-4,11)),T=0.2,logZ=w0.logZ)
        self.assertLessEqual(w2.nfev,2)

if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
import scipy.spatial.distance as distance
from bussilab.coretools import TestCase

def dataset1():
    return np.array([[-0.27434351,  0.06897978,  0.691931  ,  0.03051945, -1.11740313],
                     [ 0.09421232,  0.18027066, -0.06561068,  0.43135584, -0.43940911],
                     [-0.58521568,  0.03401204, -0.63839148, -0.23572534,  0.7680671 ],
                     [ 0.76687884,  0.62421027,  0.16370233, -0.899945  ,  0.68917803],
                     [-1.03296177, -1.64284264, -1.14054955, -0.46923336,  0.47111502],
                     [ 0.66999821, -0.48762024, -0.10909411,  0.50337187,  0.32333498],
                     [-1.5843071 , -0.77184176, -1.27124076,  0.46066538, -0.48326112],
                     [-0.34907915, -0.78747764,  1.0505906 ,  0.16638586,  1.37716574],
                     [ 0.83328731,  1.01717983,  0.10425336, -0.71342138, -1.49533519],
                     [ 0.04637906,  2.22325827, -0.33923953, -0.72341812, -0.10958039],
                     [ 1.02688336, -0.06777232, -0.54296664,  1.16736651,  1.25059246],
                     [-0.49533634, -0.26167575,  1.31880093,  1.70218493,  0.43341073],
                     [ 0.76752508, -0.27790873, -1.31025055, -2.67449964,  0.11026565],
                     [-0.46181482, -0.55080617, -1.11513591,  0.53394519, -0.22060322],
                     [-0.17277899,  0.0679858 ,  0.04163297,  0.05483895, -1.32811042],
                     [ 0.55942438, -0.45916286, -2.24796214, -0.32409745,  0.37803287],
                     [ 1.87984435, -0.52573726, -0.15719837, -0.78858451, -1.08018692],
                     [-0.25623155,  0.85477056, -1.0210906 , -1.17935635, -0.08414587],
                     [-1.19320835,  0.38967952,  1.70916784,  1.06780696, -2.24571906],
                     [ 0.44823174,  1.59230506,  0.15871062, -0.23835377,  0.59421649],
                     [-1.06998977, -0.91881022,  0.05298366, -0.97827771, -1.78408201],
                     [-0.81556567, -1.50470659, -1.38748846,  0.06731462, -1.01353324],
                     [-1.68093824, -1.00525244,  0.03008928, -0.03406712,  1.42217833],
                     [-0.81782272, -1.66081026, -0.39005011, -1.05360066,  0.74259682],
                     [ 1.29354339, -0.15176496,  1.28761518, -0.61495318,  0.18550871],
                     [ 1.55105343, -1.25720131, -0.27978451,  0.29854249,  0.17464664],
                     [-0.20532629, -0.5092539 ,  0.5594995 , -0.26360377, -0.48954807],
                     [-0.05945412, -1.66385354, -1.84581852,  0.07426876,  0.23254214],
                     [ 1.24005529,  1.2027705 ,  0.48165738,  0.7507277 ,  0.55896436],
                     [-0.50173855,  0.39622259, -1.57123979, -1.85249664,  1.4424275 ]])

def dataset1_weights():
    return np.array([1.00937788, 0.99180958, 1.00778711, 0.98892454, 1.00444056,
                     0.99750995, 1.01300244, 0.99761303, 0.98371752, 0.98464331,
                     0.99546134, 1.01373022, 0.99337099, 1.01127106, 0.99460744,
                     0.99053065, 1.01448288, 1.00005337, 1.0102084 , 0.9901373 ,
                     0.99089065, 0.99662465, 0.99441189, 0.99437434, 1.00473962,
                     1.00137577, 0.98325373, 1.01462485, 1.01370019, 0.99486246])

def dataset2():
    return np.array([[-0.59339408,  1.45733255],
                     [-0.42087971,  0.20388607],
                     [ 0.89720348, -0.53101914],
                     [-0.85465884,  0.80789103],
                     [ 0.32262436, -0.72368994]])

def dataset2_weights():
    return np.array([1.    , 1.0025, 1.005 , 1.0075, 1.01  ])

class TestClustering(TestCase):
    def test_maxclique(self):
        from bussilab.clustering import max_clique
        data=(np.array(range(50))**2).reshape(-1,1)
        dist=distance.squareform(distance.pdist(data))
        adj=np.int_(dist<1000)
        cl=max_clique(adj)
        self.assertEqual(cl.method,"max_clique")
        self.assertEqual(cl.weights,[32, 13, 5])
        ref=[range(32), range(32,45), range(45,50)]
        # compare sets, since order is irrelevant
        for i in range(len(cl.clusters)):
            self.assertEqual(set(ref[i]),set(cl.clusters[i]))

        weights=np.hstack((np.ones(32),3*np.ones(13),8*np.ones(5)))
        cl=max_clique(adj,weights)
        self.assertEqual(cl.weights,[61, 32, 18])
        ref=[range(38,50), range(0,32), range(32,38)]
        for i in range(len(cl.clusters)):
            self.assertEqual(set(ref[i]),set(cl.clusters[i]))

    def test_daura(self):
        from bussilab.clustering import daura
        data=(np.array(range(50))**2).reshape(-1,1)
        dist=distance.squareform(distance.pdist(data))
        adj=np.int_(dist<500)
        cl=daura(adj)
        self.assertEqual(adj.shape,dist.shape)  # check that adj has not been resized
        self.assertEqual(cl.method,"daura")
        self.assertEqual(cl.weights,[32, 13, 5])
        ref=[range(32), range(32,45), range(45,50)]
        # compare sets, since order is irrelevant
        for i in range(len(cl.clusters)):
            self.assertEqual(set(ref[i]),set(cl.clusters[i]))

        data=np.array([0,1,10,10,10]).reshape(-1,1)
        dist=distance.squareform(distance.pdist(data))
        adj=np.int_(dist<3)
        cl=daura(adj)
        self.assertEqual(cl.method,"daura")
        self.assertEqual(cl.weights,[3,2])
        ref=[[2,3,4],[0,1]]
        for i in range(len(cl.clusters)):
            self.assertEqual(set(ref[i]),set(cl.clusters[i]))

        data=np.array([0,1,10]).reshape(-1,1)
        weights=np.array([1,1,3])
        dist=distance.squareform(distance.pdist(data))
        adj=np.int_(dist<3)
        cl=daura(adj,weights)
        self.assertEqual(cl.method,"daura")
        self.assertEqual(cl.weights,[3,2])
        ref=[[2],[0,1]]
        for i in range(len(cl.clusters)):
            self.assertEqual(set(ref[i]),set(cl.clusters[i]))

        data=np.array([10,0,1]).reshape(-1,1)
        weights=np.array([3,1,1])
        dist=distance.squareform(distance.pdist(data))
        adj=np.int_(dist<3)
        cl=daura(adj,weights)
        self.assertEqual(adj.shape,dist.shape)  # check that adj has not been resized
        self.assertEqual(weights.shape,(len(dist),))  # check that adj has not been resized
        self.assertEqual(cl.method,"daura")
        self.assertEqual(cl.weights,[3,2])
        ref=[[0],[1,2]]
        for i in range(len(cl.clusters)):
            self.assertEqual(set(ref[i]),set(cl.clusters[i]))

    def test_qt(self):
        from bussilab.clustering import qt
        data=(np.array(range(50))**2).reshape(-1,1)
        dist=distance.squareform(distance.pdist(data))
        cl=qt(dist,1000)
        self.assertEqual(cl.method,"qt")
        self.assertEqual(cl.weights,[32, 13, 5])
        ref=[range(32), range(32,45), range(45,50)]
        # compare sets, since order is irrelevant
        for i in range(len(cl.clusters)):
            self.assertEqual(set(ref[i]),set(cl.clusters[i]))

        weights=np.hstack((np.ones(32),3*np.ones(13),8*np.ones(5)))
        cl=qt(dist,1000,weights)
        self.assertEqual(cl.weights,[61, 32, 18])
        ref=[range(38,50), range(0,32), range(32,38)]
        for i in range(len(cl.clusters)):
            self.assertEqual(set(ref[i]),set(cl.clusters[i]))

    def test_max_clique2(self):
        from bussilab.clustering import max_clique
        dist=distance.squareform(distance.pdist(dataset1()))
        # skip this since it is not reproducible on python 3.6/3.7
        #cl=max_clique(dist<3)
        #ref=[[1, 26, 2, 9, 0, 3, 8, 17, 19, 28, 14], [5, 13, 4, 23, 21, 27, 25, 15], [11, 7, 24, 10], [12, 16], [20, 18], [6, 22], [29]]
        #refw=[11, 8, 4, 2, 2, 2, 1]
        #self.assertEqual(cl.method,"max_clique")
        #self.assertEqual(cl.weights,refw)
        #for i in range(len(cl.clusters)):
        #    self.assertEqual(set(ref[i]),set(cl.clusters[i]))

        weights=dataset1_weights()
        cl=max_clique(dist<3,weights)
        ref=[[1, 15, 5, 13, 2, 23, 4, 21, 27, 17, 6], [26, 0, 14, 28, 3, 25, 16, 24], [11, 7, 10], [29, 9, 19], [20, 18], [22], [12], [8]]
        refw=[11.0220285541, 8.01046204991, 3.006804594578, 2.969643064484, 2.001099056174, 0.994411893694, 0.993370993929, 0.983717515293]
        self.assertEqual(cl.method,"max_clique")
        self.assertAlmostEqual(np.sum((cl.weights-np.array(refw))**2),0.0)
        for i in range(len(cl.clusters)):
            self.assertEqual(set(ref[i]),set(cl.clusters[i]))

    def test_max_clique3(self):
        from bussilab.clustering import max_clique
        dist=distance.squareform(distance.pdist(dataset1()))
        cl=max_clique(dist<3,min_size=4)
        ref=[[1, 26, 2, 9, 0, 3, 8, 17, 19, 28, 14], [5, 13, 4, 23, 21, 27, 25, 15], [11, 7, 24, 10]]
        refw=[11, 8, 4]
        self.assertEqual(cl.method,"max_clique")
        self.assertEqual(cl.weights,refw)
        for i in range(len(cl.clusters)):
            self.assertEqual(set(ref[i]),set(cl.clusters[i]))

        weights=dataset1_weights()
        cl=max_clique(dist<3,weights,min_size=2.96)
        ref=[[1, 15, 5, 13, 2, 23, 4, 21, 27, 17, 6], [26, 0, 14, 28, 3, 25, 16, 24], [11, 7, 10], [29, 9, 19]]
        refw=[11.0220285541, 8.01046204991, 3.006804594578, 2.969643064484]
        self.assertEqual(cl.method,"max_clique")
        self.assertAlmostEqual(np.sum((cl.weights-np.array(refw))**2),0.0)
        for i in range(len(cl.clusters)):
            self.assertEqual(set(ref[i]),set(cl.clusters[i]))

    def test_max_clique4(self):
        from bussilab.clustering import max_clique
        dist=distance.squareform(distance.pdist(dataset1()))
        cl=max_clique(dist<3,max_clusters=2)
        ref=[[1, 26, 2, 9, 0, 3, 8, 17, 19, 28, 14], [5, 13, 4, 23, 21, 27, 25, 15]]
        refw=[11, 8]
        self.assertEqual(cl.method,"max_clique")
        self.assertEqual(cl.weights,refw)
        for i in range(len(cl.clusters)):
            self.assertEqual(set(ref[i]),set(cl.clusters[i]))

        weights=dataset1_weights()
        cl=max_clique(dist<3,weights,max_clusters=2)
        ref=[[1, 15, 5, 13, 2, 23, 4, 21, 27, 17, 6], [26, 0, 14, 28, 3, 25, 16, 24]]
        refw=[11.0220285541, 8.01046204991]
        self.assertEqual(cl.method,"max_clique")
        self.assertAlmostEqual(np.sum((cl.weights-np.array(refw))**2),0.0)
        for i in range(len(cl.clusters)):
            self.assertEqual(set(ref[i]),set(cl.clusters[i]))

    def test_qt2(self):
        from bussilab.clustering import qt
        dist=distance.squareform(distance.pdist(dataset1()))
        dist[0,0]=1e-3 # this is to check that qt() ignores the diagonal terms
        cl=qt(dist,3)
        ref=[[13, 6, 21, 4, 27, 23, 2, 5, 15, 1, 17], [3, 19, 28, 24, 8, 0, 26, 14], [11, 7, 10], [16, 25], [29, 12], [18, 20], [9], [22]]
        refw=[11, 8, 3, 2, 2, 2, 1, 1]
        self.assertEqual(cl.method,"qt")
        self.assertEqual(cl.weights,refw)
        for i in range(len(cl.clusters)):
            self.assertEqual(set(ref[i]),set(cl.clusters[i]))

        weights=dataset1_weights()
        cl=qt(dist,3,weights)
        ref=[[13, 6, 21, 4, 27, 23, 2, 5, 15, 1, 17], [16, 25, 24, 26, 3, 14, 0, 28], [11, 7, 10], [19, 9, 8], [18, 20], [29, 12], [22]]
        refw=[11.02202855414899, 8.010462049910068, 3.0068045945788784, 2.958498123817103, 2.0010990561741044, 1.9882334498903202, 0.99441189]
        self.assertEqual(cl.method,"qt")
        self.assertAlmostEqual(np.sum((cl.weights-np.array(refw))**2),0.0)
        for i in range(len(cl.clusters)):
            self.assertEqual(set(ref[i]),set(cl.clusters[i]))

    def test_qt3(self):
        from bussilab.clustering import qt
        dist=distance.squareform(distance.pdist(dataset1()))
        cl=qt(dist,3,min_size=4)
        ref=[[13, 6, 21, 4, 27, 23, 2, 5, 15, 1, 17], [3, 19, 28, 24, 8, 0, 26, 14]]
        refw=[11, 8]
        self.assertEqual(cl.method,"qt")
        self.assertEqual(cl.weights,refw)
        for i in range(len(cl.clusters)):
            self.assertEqual(set(ref[i]),set(cl.clusters[i]))

        weights=dataset1_weights()
        cl=qt(dist,3,weights,min_size=2.96)
        ref=[[13, 6, 21, 4, 27, 23, 2, 5, 15, 1, 17], [16, 25, 24, 26, 3, 14, 0, 28], [11, 7, 10]]
        refw=[11.02202855414899, 8.010462049910068, 3.0068045945788784]
        self.assertEqual(cl.method,"qt")
        self.assertAlmostEqual(np.sum((cl.weights-np.array(refw))**2),0.0)
        for i in range(len(cl.clusters)):
            self.assertEqual(set(ref[i]),set(cl.clusters[i]))

    def test_qt4(self):
        from bussilab.clustering import qt
        dist=distance.squareform(distance.pdist(dataset1()))
        cl=qt(dist,3,max_clusters=2)
        ref=[[13, 6, 21, 4, 27, 23, 2, 5, 15, 1, 17], [3, 19, 28, 24, 8, 0, 26, 14]]
        refw=[11, 8]
        self.assertEqual(cl.method,"qt")
        self.assertEqual(cl.weights,refw)
        for i in range(len(cl.clusters)):
            self.assertEqual(set(ref[i]),set(cl.clusters[i]))

        weights=dataset1_weights()
        cl=qt(dist,3,weights,max_clusters=2)
        ref=[[13, 6, 21, 4, 27, 23, 2, 5, 15, 1, 17], [16, 25, 24, 26, 3, 14, 0, 28]]
        refw=[11.0220285541, 8.01046204991]
        self.assertEqual(cl.method,"qt")
        self.assertAlmostEqual(np.sum((cl.weights-np.array(refw))**2),0.0)
        for i in range(len(cl.clusters)):
            self.assertEqual(set(ref[i]),set(cl.clusters[i]))
    def test_qt5(self):
        from bussilab.clustering import qt
        # test with identical points
        dist=distance.squareform(distance.pdist([[1],[1],[1],[2],[2],[3]]))
        cl=qt(dist,0.5)
        ref=[[0,1,2],[3,4],[5],[6]]
        refw=[3,2,1]
        self.assertEqual(cl.method,"qt")
        self.assertEqual(cl.weights,refw)
        for i in range(len(cl.clusters)):
            self.assertEqual(set(ref[i]),set(cl.clusters[i]))

        dist=distance.squareform(distance.pdist([[1],[1],[1],[2],[2],[3],[4],[5]]))
        weights=np.array([1,1,1,2,2,5,0.5,0.1])
        cl=qt(dist,0.5,weights)
        ref=[[5],[3,4],[0,1,2],[6],[7]]
        refw=[5,4,3,0.5,0.1]
        self.assertEqual(cl.method,"qt")
        self.assertEqual(cl.weights,refw)
        for i in range(len(cl.clusters)):
            self.assertEqual(set(ref[i]),set(cl.clusters[i]))

    def test_qt6(self):
        from bussilab.clustering import qt
        dist=distance.squareform(distance.pdist(dataset2()))
        weights=dataset2_weights()
        cl=qt(dist,1.0,weights)
        ref=[[2, 4],[1, 3],[0]]
        refw=[2.015, 2.01, 1.0]
        self.assertEqual(cl.method,"qt")
        self.assertAlmostEqual(np.sum((cl.weights-np.array(refw))**2),0.0)
        for i in range(len(cl.clusters)):
            self.assertEqual(set(ref[i]),set(cl.clusters[i]))

    def test_qt7(self):
        from bussilab.clustering import qt
        data=np.array([[1],[2.1],[10],[11]])
        dist=distance.squareform(distance.pdist(data))
        cl=qt(dist,2.0)
        ref=[[2, 3],[0, 1]]
        refw=[2,2]
        self.assertEqual(cl.method,"qt")
        self.assertEqual(cl.weights,refw)
        self.assertAlmostEqual(np.sum((cl.weights-np.array(refw))**2),0.0)
        for i in range(len(cl.clusters)):
            self.assertEqual(set(ref[i]),set(cl.clusters[i]))

    def test_qt8(self):
        from bussilab.clustering import qt
        # this is to test that [2,3] gets priority wrt [0,1] since it is more compact
        dist=distance.squareform(distance.pdist(np.array([[0.0],[0.5],[1.0],[1.2],])))
        weights=np.array([1,1,1,1])
        cl=qt(dist,0.6,weights)
        ref=[[2,3],[0,1]]
        refw=[2,2]
        self.assertEqual(cl.method,"qt")
        self.assertAlmostEqual(np.sum((cl.weights-np.array(refw))**2),0.0)
        for i in range(len(cl.clusters)):
            self.assertEqual(set(ref[i]),set(cl.clusters[i]))

    def test_qt8(self):
        from bussilab.clustering import qt
        # this is to test that [1] grows with [2] rather than with [0] since [2] has
        # more weight than [0].
        # notice that [2] grows with [3], which is closer
        # however, [1,2] has weight 2.15 and is preferred over [2,3], which has weight 2.1
        dist=distance.squareform(distance.pdist(np.array([[0.0],[1.0],[2.0],[2.5]])))
        weights=np.array([1,1.05,1.1,1])
        cl=qt(dist,1.2,weights)
        ref=[[1,2],[0],[3]]
        refw=[2.15,1,1]
        self.assertEqual(cl.method,"qt")
        self.assertAlmostEqual(np.sum((cl.weights-np.array(refw))**2),0.0)
        for i in range(len(cl.clusters)):
            self.assertEqual(set(ref[i]),set(cl.clusters[i]))


try:
    import networkit
    _has_networkit=True
except ModuleNotFoundError:
    _has_networkit=False

if _has_networkit:
    class TestClusteringNetworkit(TestCase):
        def test_maxclique(self):
            from bussilab.clustering import max_clique
            data=(np.array(range(50))**2).reshape(-1,1)
            dist=distance.squareform(distance.pdist(data))
            adj=np.int_(dist<1000)
            cl=max_clique(adj,use_networkit=True)
            self.assertEqual(cl.method,"max_clique")
            self.assertEqual(cl.weights,[32, 13, 3, 2])
            ref=[range(32), range(34,47), range(47,50),(32,33)]
            # compare sets, since order is irrelevant
            for i in range(len(cl.clusters)):
                self.assertEqual(set(ref[i]),set(cl.clusters[i]))

            weights=np.hstack((np.ones(32),3*np.ones(13),8*np.ones(5)))
            cl=max_clique(adj,weights,use_networkit=True)
            self.assertEqual(cl.weights,[61, 32, 18])
            ref=[range(38,50), range(0,32), range(32,38)]
            for i in range(len(cl.clusters)):
                self.assertEqual(set(ref[i]),set(cl.clusters[i]))


if __name__ == "__main__":
    unittest.main()




import os
import unittest
import time

from bussilab.cli import cli
from bussilab.coretools import TestCase
from bussilab.coretools import cd

class TestCron(TestCase):
    def test_cron(self):
        with cd(os.path.dirname(os.path.abspath(__file__))):
            try:
                os.remove("cron.out")
            except FileNotFoundError:
                pass
            try:
                os.remove("cron_screen.out")
            except FileNotFoundError:
                pass
            try:
                os.remove("cron_screen2.out")
            except FileNotFoundError:
                pass
            try:
                os.remove("screenlog.0")
            except FileNotFoundError:
                pass
            try:
                os.remove("cron_reboot.out")
            except FileNotFoundError:
                pass
            try:
                os.remove("cron_reboot_unsorted.out")
            except FileNotFoundError:
                pass
            # this is asynchronous
            cli("cron --cron-file cron_screen.yml --period 2 --max-times 2 --detach --screen-log screenlog.0 --unique")
            # the second instance will not be run (--unique)
            cli("cron --cron-file cron_screen.yml --period 2 --max-times 2 --detach --screen-log screenlog.0 --unique")
            # also this is asynchronoous
            cli("cron --cron-file cron_reboot.yml --max-times 4 --detach --screen-log screenlog.0")
            now=time.time()
            # this is synchronous, and will take some time
            cli("cron --cron-file cron.yml --max-times 2 --no-screen")
            # make sure to give enough time to the asynchronous test
            time.sleep(9-(time.time()-now))
            # this is for debugging, should be shown only if there is an error later
            os.system("ls -ltr")
            os.system("echo == screenlog ==")
            os.system("cat screenlog.0")
            os.system("echo == end screenlog ==")
            self.assertEqualFile("cron.out")
            self.assertEqualFile("cron_screen.out")
            self.assertEqualFile("cron_screen2.out")
            self.assertEqualFile("cron_reboot.out")
            with open("cron_reboot_unsorted.out") as file:
                lines = [line.rstrip() for line in file]
                lines.sort()
                self.assertEqual(lines,["C0","C0","C1","C1"])
            os.remove("cron.out")
            os.remove("cron_screen.out")
            os.remove("cron_screen2.out")
            os.remove("cron_reboot.out")
            os.remove("cron_reboot_unsorted.out")
            os.remove("screenlog.0")

if __name__ == "__main__":
    unittest.main()



import unittest

from bussilab import potts
from bussilab.coretools import TestCase

class TestPotts(TestCase):
    def run_potts(self,fullmatrix):
        import numpy as np
        m=potts.Model(3,fullmatrix=fullmatrix)
        h=np.array((-1.0,0.0,1.0))
        J=np.array(((0.0,0.3,-0.2),
                    (0.3,0.0,-0.6),
                    (-0.2,-0.6,0.0)))
        averages=m.compute(h,J)[1]
        infer=m.infer(averages)
        h_,J_=infer.h,infer.J
        print(averages)
        print(infer)
        self.assertTrue(infer.success)
        self.assertAlmostEqual(np.sum((h-h_)**2),0.0)
        self.assertAlmostEqual(np.sum((J-J_)**2),0.0)

# Temporarily removed:
    def test_potts1(self):
        self.run_potts(True)

# Temporarily removed:
    def test_potts2(self):
        self.run_potts(False)

if __name__ == "__main__":
    unittest.main()



import unittest
import numpy as np
from bussilab.lohman import lohman
from bussilab.coretools import TestCase

class TestLohman(TestCase):
    def test_lohman(self):
        t = np.linspace(0, 4, 5)
        a = lohman(t, ku=1, kd=0.1)
        aref = np.array((0.0, 0.60648083, 0.80836077, 0.87556076, 0.89792969))
        b = lohman(t, ku=2, kd=0.1)
        bref = np.array((0.0, 0.83575578, 0.93809945, 0.95063209, 0.95216679))
        c = lohman(t, ku=1, kd=0.2)
        cref = np.array((0.0, 0.58233816, 0.75773504, 0.81056356, 0.82647521))
        d = lohman(t, ku=1, kd=0.1, n=2)
        dref = np.array((0.0, 0.24873614, 0.53341313, 0.69537387, 0.77165485))
        e = lohman(t, ku=1, kd=0.1, boundaries=(0.3, 0.7))
        eref = np.array((0.3, 0.54259233, 0.62334431, 0.6502243, 0.65917188))
        self.assertAlmostEqual(np.sum((a-aref)**2), 0.0)
        self.assertAlmostEqual(np.sum((b-bref)**2), 0.0)
        self.assertAlmostEqual(np.sum((c-cref)**2), 0.0)
        self.assertAlmostEqual(np.sum((d-dref)**2), 0.0)
        self.assertAlmostEqual(np.sum((e-eref)**2), 0.0)

if __name__ == "__main__":
    unittest.main()



import os
import unittest

from bussilab.notify import notify

# only run tests if env vars are configured
if 'BUSSILAB_TEST_NOTIFY_TOKEN' in os.environ:
    token=os.environ["BUSSILAB_TEST_NOTIFY_TOKEN"]
    channel=os.environ["BUSSILAB_TEST_NOTIFY_CHANNEL"]
    class TestNotify(unittest.TestCase):
        def test_notify(self):
            url=notify("unittest1", token=token, channel=channel)
            notify(react=url+":white_check_mark",token=token)
            notify(delete=url+":white_check_mark",token=token)

            url=notify("unittest2 *WRONG*", token=token, channel=channel)
            notify("unittest2", update=url, token=token)

            url=notify("unittest3 *WRONG*", token=token, channel=channel)
            url=notify("unittest3 *WRONG 2*", update=url, token=token)
            url2=notify("unittest3 reply", reply=url, token=token)
            url3=notify("unittest3 broadcast", reply_broadcast=url, token=token)
            notify(delete=url, token=token)

            with self.assertRaises(Exception):
                notify("unittest3 *WRONG 3*", update=url, token=token)

            with self.assertRaises(Exception):
                notify(delete=url, token=token)

            with self.assertRaises(TypeError):
                notify("unittest1", token=token, channel=channel, update=url)
            with self.assertRaises(TypeError):
                notify("unittest1", token=token, channel=channel, delete=url)
            with self.assertRaises(TypeError):
                notify("unittest1", token=token, update=url, delete=url)
            with self.assertRaises(TypeError):
                notify("unittest1", token=token, update="incorrect-url")
            with self.assertRaises(TypeError):
                notify("unittest1", token=token, delete="incorrect-url")

            url=notify("unittest4 *WRONG*", token=token, channel=channel)
            url2=notify("test upload",file=os.path.realpath(__file__), token=token, reply=url)
            notify(delete=url, token=token)
            notify(delete=url2, token=token)
            
            url=notify("test upload",file=os.path.realpath(__file__), token=token, channel=channel)
            notify(delete=url, token=token)


if __name__ == "__main__":
    unittest.main()



import unittest

from bussilab.maxent import maxent

class TestMaxent(unittest.TestCase):
    def test_maxent(self):
        import numpy as np
        traj = [[2], [4]]

        m = maxent(traj, [2.5])
        self.assertTrue(m.success)
        self.assertAlmostEqual(np.sum((m.logW_ME-[-0.28768, -1.38630])**2), 0.0)
        self.assertAlmostEqual((m.averages[0]-2.5)**2, 0.0)
        self.assertAlmostEqual((m.lambdas[0]-0.549306)**2, 0.0)

        m = maxent(traj, [2.5], logW=[0, -5])
        self.assertTrue(m.success)
        self.assertAlmostEqual(np.sum((m.logW_ME-[-0.28768, -1.38630])**2), 0.0)
        self.assertAlmostEqual((m.averages[0]-2.5)**2, 0.0)
        self.assertAlmostEqual((m.lambdas[0]- -1.95069378)**2, 0.0)

        m = maxent(traj, [2.5], logW=[0, -5], l2=1)
        self.assertTrue(m.success)
        self.assertAlmostEqual(np.sum((m.logW_ME-[-0.01697795, -4.08431654])**2), 0.0)
        self.assertAlmostEqual((m.averages[0]-2.03367)**2, 0.0)
        self.assertAlmostEqual((m.lambdas[0]- -0.46633)**2, 0.0)

        def reg(x):
            return 0.5*0.5*np.sum(x**2), 0.5*x
        m = maxent(traj, [2.5], logW=[0, -5], l2=[0.5], regularization=reg)
        self.assertTrue(m.success)
        self.assertAlmostEqual(np.sum((m.logW_ME-[-0.01697795, -4.08431654])**2), 0.0)
        self.assertAlmostEqual((m.averages[0]-2.03367)**2, 0.0)
        self.assertAlmostEqual((m.lambdas[0]- -0.46633)**2, 0.0)

    def test_maxent2d(self):
        import numpy as np
        traj = [[0, 0], [1, 0], [0, 1]]

        m = maxent(traj, (0.3, 0.4))
        self.assertTrue(m.success)
        self.assertAlmostEqual(np.sum((m.logW_ME-[-1.20396686, -1.20398582, -0.91628543])**2), 0.0)
        self.assertAlmostEqual(np.sum((m.averages-[0.3, 0.4])**2), 0.0)
        self.assertAlmostEqual(np.sum((m.lambdas-[1.89607144e-05, -2.87681431e-01])**2), 0.0)

        m = maxent(traj, ((-np.inf, 0.3), (0.4, np.inf)))
        self.assertTrue(m.success)
        self.assertAlmostEqual(np.sum((m.logW_ME-[-1.20396686, -1.20398582, -0.91628543])**2), 0.0)
        self.assertAlmostEqual(np.sum((m.averages-[0.3, 0.4])**2), 0.0)
        self.assertAlmostEqual(np.sum((m.lambdas-[1.89607144e-05, -2.87681431e-01])**2), 0.0)

        m = maxent(traj, ((-np.inf, 0.3), 0.4))
        self.assertTrue(m.success)
        self.assertAlmostEqual(np.sum((m.logW_ME-[-1.20396686, -1.20398582, -0.91628543])**2), 0.0)
        self.assertAlmostEqual(np.sum((m.averages-[0.3, 0.4])**2), 0.0)
        self.assertAlmostEqual(np.sum((m.lambdas-[1.89607144e-05, -2.87681431e-01])**2), 0.0)

        m = maxent(traj, ((-np.inf, 0.3), 0.4), logW=(0, 1, 2))
        self.assertTrue(m.success)
        self.assertAlmostEqual(np.sum((m.logW_ME-[-1.20396686, -1.20398582, -0.91628543])**2), 0.0)
        self.assertAlmostEqual(np.sum((m.averages-[0.3, 0.4])**2), 0.0)
        self.assertAlmostEqual(np.sum((m.lambdas-[0.99998443, 1.71233637])**2), 0.0)

        m = maxent(traj, ((0.3, np.inf), (-np.inf, 0.4)))
        self.assertTrue(m.success)
        self.assertAlmostEqual(np.sum((m.logW_ME-[-1.0986122, -1.0986122, -1.0986122])**2), 0.0)
        self.assertAlmostEqual(np.sum((m.averages-[0.33333333, 0.333333333])**2), 0.0)
        self.assertAlmostEqual(np.sum((m.lambdas-[0.0, 0.0])**2), 0.0)

        m = maxent(traj, ((0.3, 1.0), (0.0, 0.4)))
        self.assertTrue(m.success)
        self.assertAlmostEqual(np.sum((m.logW_ME-[-1.0986122, -1.0986122, -1.0986122])**2), 0.0)
        self.assertAlmostEqual(np.sum((m.averages-[0.33333333, 0.333333333])**2), 0.0)
        self.assertAlmostEqual(np.sum((m.lambdas-[0.0, 0.0])**2), 0.0)

        m = maxent(traj, ((0.3, 1.0), (-np.inf, 0.4)))
        self.assertTrue(m.success)
        self.assertAlmostEqual(np.sum((m.logW_ME-[-1.0986122, -1.0986122, -1.0986122])**2), 0.0)
        self.assertAlmostEqual(np.sum((m.averages-[0.33333333, 0.333333333])**2), 0.0)
        self.assertAlmostEqual(np.sum((m.lambdas-[0.0, 0.0])**2), 0.0)

try:
    import cudamat
    _has_cudamat=True
except ModuleNotFoundError:
    _has_cudamat=False

if _has_cudamat:
    class TestCuda(unittest.TestCase):
        def test(self):
            import numpy as np
            traj = [[0, 0], [1, 0], [0, 1]]
            m = maxent(traj, ((0.3, 1.0), (-np.inf, 0.4)),cuda=True)
            self.assertTrue(m.success)
            self.assertAlmostEqual(np.sum((m.logW_ME-[-1.0986122, -1.0986122, -1.0986122])**2), 0.0)
            self.assertAlmostEqual(np.sum((m.averages-[0.33333333, 0.333333333])**2), 0.0)
            self.assertAlmostEqual(np.sum((m.lambdas-[0.0, 0.0])**2), 0.0)

            cu_traj=cudamat.CUDAMatrix(np.array(traj))
            m = maxent(cu_traj, ((0.3, 1.0), (-np.inf, 0.4)),cuda=True)
            self.assertTrue(m.success)
            self.assertAlmostEqual(np.sum((m.logW_ME-[-1.0986122, -1.0986122, -1.0986122])**2), 0.0)
            self.assertAlmostEqual(np.sum((m.averages-[0.33333333, 0.333333333])**2), 0.0)
            self.assertAlmostEqual(np.sum((m.lambdas-[0.0, 0.0])**2), 0.0)

if __name__ == "__main__":
    unittest.main()



import unittest

import bussilab.coretools as coretools

class Test(unittest.TestCase):
    def test_ensure_np_array(self):
        import numpy as np

        # None
        A = None
        B = coretools.ensure_np_array(A)
        self.assertIsNone(B)

        # Tuple
        A = (1.0, 2.0, 3.0)
        B = coretools.ensure_np_array(A)
        self.assertIsInstance(B, np.ndarray)
        self.assertSequenceEqual(list(B), list(A))
        # Check that it is a deep copy
        B[1] = 10.0
        self.assertSequenceEqual(A, [1, 2, 3])

        # List
        A = [1.0, 2.0, 3.0]
        B = coretools.ensure_np_array(A)
        self.assertIsInstance(B, np.ndarray)
        self.assertSequenceEqual(list(B), list(A))
        # Check that it is a deep copy
        B[1] = 10.0
        self.assertSequenceEqual(A, [1, 2, 3])

        # Np array
        A = np.array([1.0, 2.0, 3.0])
        B = coretools.ensure_np_array(A)
        self.assertIsInstance(B, np.ndarray)
        self.assertSequenceEqual(list(B), list(A))
        # Check that it is a shallow copy
        B[1] = 10.0
        self.assertSequenceEqual(list(A), list(B))

        # Rank2 tuple
        A = ((1.0, 2.0, 3.0), (4.0, 5.0, 6.0))
        B = coretools.ensure_np_array(A)
        self.assertIsInstance(B, np.ndarray)
        self.assertAlmostEqual((B[0, 0]-1)**2 + (B[0, 1]-2)**2 + (B[0, 2]-3)**2, 0.0)
        self.assertAlmostEqual((B[1, 0]-4)**2 + (B[1, 1]-5)**2 + (B[1, 2]-6)**2, 0.0)
        # Check that it is a deep copy
        B[1, 1] = 10.0
        self.assertAlmostEqual((A[0][0]-1)**2 + (A[0][1]-2)**2 + (A[0][2]-3)**2, 0.0)
        self.assertAlmostEqual((A[1][0]-4)**2 + (A[1][1]-5)**2 + (A[1][2]-6)**2, 0.0)

        # Rank2 list
        A = [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]]
        B = coretools.ensure_np_array(A)
        self.assertIsInstance(B, np.ndarray)
        self.assertAlmostEqual((B[0, 0]-1)**2 + (B[0, 1]-2)**2 + (B[0, 2]-3)**2, 0.0)
        self.assertAlmostEqual((B[1, 0]-4)**2 + (B[1, 1]-5)**2 + (B[1, 2]-6)**2, 0.0)
        # Check that it is a deep copy
        B[1, 1] = 10.0
        self.assertAlmostEqual((A[0][0]-1)**2 + (A[0][1]-2)**2 + (A[0][2]-3)**2, 0.0)
        self.assertAlmostEqual((A[1][0]-4)**2 + (A[1][1]-5)**2 + (A[1][2]-6)**2, 0.0)

        # Tuple of np arrays
        A = (np.array((1.0, 2.0, 3.0)), np.array((4.0, 5.0, 6.0)))
        B = coretools.ensure_np_array(A)
        self.assertIsInstance(B, np.ndarray)
        self.assertAlmostEqual((B[0, 0]-1)**2 + (B[0, 1]-2)**2 + (B[0, 2]-3)**2, 0.0)
        self.assertAlmostEqual((B[1, 0]-4)**2 + (B[1, 1]-5)**2 + (B[1, 2]-6)**2, 0.0)
        # Check that it is a deep copy
        B[1, 1] = 10.0
        self.assertAlmostEqual((A[0][0]-1)**2 + (A[0][1]-2)**2 + (A[0][2]-3)**2, 0.0)
        self.assertAlmostEqual((A[1][0]-4)**2 + (A[1][1]-5)**2 + (A[1][2]-6)**2, 0.0)

        # List of np arrays
        A = [np.array((1.0, 2.0, 3.0)), np.array((4.0, 5.0, 6.0))]
        B = coretools.ensure_np_array(A)
        self.assertIsInstance(B, np.ndarray)
        self.assertAlmostEqual((B[0, 0]-1)**2 + (B[0, 1]-2)**2 + (B[0, 2]-3)**2, 0.0)
        self.assertAlmostEqual((B[1, 0]-4)**2 + (B[1, 1]-5)**2 + (B[1, 2]-6)**2, 0.0)
        # Check that it is a deep copy
        B[1, 1] = 10.0
        self.assertAlmostEqual((A[0][0]-1)**2 + (A[0][1]-2)**2 + (A[0][2]-3)**2, 0.0)
        self.assertAlmostEqual((A[1][0]-4)**2 + (A[1][1]-5)**2 + (A[1][2]-6)**2, 0.0)

        # Rank2 np array
        A = np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]])
        B = coretools.ensure_np_array(A)
        self.assertIsInstance(B, np.ndarray)
        self.assertAlmostEqual((B[0, 0]-1)**2 + (B[0, 1]-2)**2 + (B[0, 2]-3)**2, 0.0)
        self.assertAlmostEqual((B[1, 0]-4)**2 + (B[1, 1]-5)**2 + (B[1, 2]-6)**2, 0.0)
        # Check that it is a shallow copy
        B[1, 1] = 10.0
        self.assertAlmostEqual((A[0][0]-1)**2 + (A[0][1]-2)**2 + (A[0][2]-3)**2, 0.0)
        self.assertAlmostEqual((A[1][0]-4)**2 + (A[1][1]-10)**2 + (A[1][2]-6)**2, 0.0)

    def test_results(self):

        class Result(coretools.Result):
            pass

        res = Result(a=1, b="test")
        self.assertEqual(res.a, 1)
        self.assertEqual(res.b, "test")
        self.assertEqual(str(res), " a: 1\n b: 'test'")

        res2 = Result(a=2, b="test2", c=res)
        self.assertEqual(res2.a, 2)
        self.assertEqual(res2.b, "test2")
        self.assertEqual(str(res2), " a: 2\n b: 'test2'\n c:  a: 1\n     b: 'test'")
        self.assertEqual(res2.c.a, 1)
        self.assertEqual(res2.c.b, "test")
        self.assertEqual(str(res2.c), " a: 1\n b: 'test'")

    def test_results2(self):
        class Result(coretools.Result):
            def __init__(self,*,val1,val2):
                self.val2=val2
                self.val1=val1
        r=Result(val1=1,val2=2)
        print(r.val1)
        print(r["val1"])
        print(r.val2)
        print(r["val2"])
        self.assertEqual(str(r)," val1: 1\n val2: 2")
        self.assertEqual(dir(r),["val1","val2"])

        del r.val1
        self.assertEqual(str(r)," val2: 2")
        self.assertEqual(dir(r),["val2"])

        with self.assertRaises(KeyError):
            del r.val1

        with self.assertRaises(AttributeError):
            print(r.attribute)
        with self.assertRaises(KeyError):
            print(r['attribute'])

        class Empty(coretools.Result):
            pass
        e=Empty()
        self.assertEqual(str(e),"Empty()")

    def test_submodules(self):
        import bussilab
        with self.assertRaises(ModuleNotFoundError):
            import bussilab.not_existing_module
        with self.assertRaises(AttributeError):
            m = bussilab.not_existing_module
        self.assertIsInstance(bussilab.describe_submodule(""),str)
        self.assertIsInstance(bussilab.describe_submodule("coretools"),str)
        with self.assertRaises(ModuleNotFoundError):
            bussilab.describe_submodule("not_existing_module")

        self.assertIsInstance(bussilab.list_submodules(),list)

    def test_cd(self):
        import bussilab
        import os
        class test_raise(Exception):
            pass
        with coretools.cd(os.path.dirname(os.path.abspath(__file__))):
            try:
                with coretools.cd("test_dir",create=True):
                    with open("test_file","w") as f:
                        print("content",file=f)
                raise test_raise # test if directory is changed back
            except test_raise:
                pass
            with open("test_dir/test_file") as f:
                for l in f:
                    self.assertEqual(l,"content\n")
            os.remove("test_dir/test_file")
            os.rmdir("test_dir")

    def test_cd2(self):
        import bussilab
        import os
        class test_raise(Exception):
            pass
        with coretools.cd(os.path.dirname(os.path.abspath(__file__))):
            try:
                with coretools.cd("test_dir/subdir",create=True):
                    with open("test_file","w") as f:
                        print("content",file=f)
                raise test_raise # test if directory is changed back
            except test_raise:
                pass
            with open("test_dir/subdir/test_file") as f:
                for l in f:
                    self.assertEqual(l,"content\n")
            os.remove("test_dir/subdir/test_file")
            os.rmdir("test_dir/subdir")
            os.rmdir("test_dir")

if __name__ == "__main__":
    unittest.main()



import unittest

import os
import numpy as np

from bussilab.ann import ANN
from bussilab.coretools import TestCase
from bussilab.coretools import cd

def derivatives(ann,n=100,prefactor=1e-10,vector=True):
    p=[]
    x=np.random.normal(size=ann.narg)
    start=np.random.normal(size=ann.npar)
    ann.setpar(start)
    gp=ann.getpar()
    assert np.all(np.equal(gp,start))
    if vector:
        v0,deriv=ann.derparVec(x.reshape((1,-1)))
    else:
        v0,deriv=ann.derpar(x)
    for i in range(n):
        test=prefactor*np.random.normal(size=ann.npar)
        ann.setpar(start+test)
        v1=ann.apply(x)
        p.append((v1-v0,np.matmul(deriv,test)))
    p=np.array(p)/prefactor
    return np.average((p[:,0]-p[:,1])**2)

class TestANN(TestCase):
    def test_ann1(self):
        np.random.seed(1977)
        self.assertLess(derivatives(ANN([10],cuda=False)),1e-9)
    def test_ann2(self):
        np.random.seed(1977)
        self.assertLess(derivatives(ANN([10,10],cuda=False)),1e-9)
    def test_ann3(self):
        np.random.seed(1977)
        self.assertLess(derivatives(ANN([10,10,10],cuda=False)),1e-8)
    def test_ann4(self):
        np.random.seed(1977)
        self.assertLess(derivatives(ANN([10,8,6,4,2],cuda=False)),1e-8)

    def test_ann1r(self):
        np.random.seed(1977)
        self.assertLess(derivatives(ANN([10],cuda=False,activation='relu')),1e-9)
    def test_ann2r(self):
        np.random.seed(1977)
        self.assertLess(derivatives(ANN([10,10],cuda=False,activation='relu')),1e-9)
    def test_ann3r(self):
        np.random.seed(1977)
        self.assertLess(derivatives(ANN([10,10,10],cuda=False,activation='relu')),1e-8)
    def test_ann4r(self):
        np.random.seed(1977)
        self.assertLess(derivatives(ANN([10,8,6,4,2],cuda=False,activation='relu')),1e-8)

    def test_small(self):
        ann=ANN([2,1],cuda=False)
        ann.setpar(np.array([1.0]*ann.npar))
        d=ann.deriv(np.array([1.0]*2))
        self.assertAlmostEqual(d[0],4.048587351573742)
        self.assertAlmostEqual(d[1][0][0,0],0.9525741268224333)
        self.assertAlmostEqual(d[1][1][0],3.048587351573742)
        self.assertAlmostEqual(d[2][0][0],0.9525741268224333)
        self.assertAlmostEqual(d[2][1][0],1.0)

    def test_small2(self):
        ann=ANN([3,2],activation="relu",cuda=False)
        ann.setpar(0.1*np.array(range(ann.npar)))

        a=ann.apply((1,2,3))
        self.assertIsInstance(a,float)
        self.assertAlmostEqual(a,4.61)
        aa=ann.apply([(1,2,3),(4,5,6)])
        self.assertIsInstance(aa,np.ndarray)
        self.assertEqual(aa.shape,(2,))
        self.assertAlmostEqual(aa[0],4.61)
        self.assertAlmostEqual(aa[1],7.58)

        d=ann.deriv((1,2,3))
        self.assertIsInstance(d,tuple)
        self.assertEqual(len(d),3)
        self.assertIsInstance(d[0],float)
        self.assertAlmostEqual(d[0],4.61)
        self.assertIsInstance(d[1],list)
        self.assertEqual(len(d[1]),2)
        self.assertEqual(d[1][0].shape,(3,2))
        self.assertAlmostEqual(d[1][0][0,0],0.6)
        self.assertAlmostEqual(d[1][0][0,1],0.7)
        self.assertAlmostEqual(d[1][0][1,0],1.2)
        self.assertAlmostEqual(d[1][0][1,1],1.4)
        self.assertAlmostEqual(d[1][0][2,0],1.8)
        self.assertAlmostEqual(d[1][0][2,1],2.1)
        self.assertEqual(d[1][1].shape,(2,1))
        self.assertAlmostEqual(d[1][1][0,0],2.4)
        self.assertAlmostEqual(d[1][1][1,0],3.1)
        self.assertIsInstance(d[2],list)
        self.assertEqual(len(d[2]),2)
        self.assertEqual(d[2][0].shape,(2,))
        self.assertAlmostEqual(d[2][0][0],0.6)
        self.assertAlmostEqual(d[2][0][1],0.7)
        self.assertEqual(d[2][1].shape,(1,))
        self.assertAlmostEqual(d[2][1][0],1.0)
        self.assertAlmostEqual(d[2][0][1],0.7)

    def backprop(self,layers=None,activation="softplus"):
        if layers is None:
            layers=[10,8,6,4,2]
        np.random.seed(1977)
        ann=ANN(layers,activation=activation,random_weights=True)
        traj=np.random.normal(size=(1000,ann.narg))
        d=ann.derpar(traj)
        reference=np.matmul(d[0],d[1])
        state=ann.forward(traj)
        der=ann.backward_par(state.f,state)
        self.assertAlmostEqual(np.sum((reference-der)**2),0.0)

    def test_backprop1(self):
        self.backprop([10,10,10,10],"softplus")
    def test_backprop2(self):
        self.backprop([10,8,6,4,2],"softplus")
    def test_backprop1r(self):
        self.backprop([10,10,10,10],"relu")
    def test_backprop2r(self):
        self.backprop([10,8,6,4,2],"relu")

    def test_plumed(self):
        ann=ANN([3,2],cuda=False)
        ann.setpar(np.array(range(ann.npar)))
        with cd(os.path.dirname(os.path.abspath(__file__))):
            ann.dumpPlumed("ann_plumed.dat","ann")
            ann.dumpPlumed("ann_plumed_combine.dat","combine","pref","x1,x2,x3")
            self.assertEqualFile("ann_plumed.dat")
            self.assertEqualFile("ann_plumed_combine.dat")
            os.remove("ann_plumed.dat")
            os.remove("ann_plumed_combine.dat")

try:
    import cudamat
    _has_cudamat=True
except ModuleNotFoundError:
    _has_cudamat=False

if _has_cudamat:
    class TestCuda(unittest.TestCase):
        def _test_layers(self,layers,activation='softplus'):
            np.random.seed(1977)
            x=np.random.normal(size=layers[0])
            a=ANN(layers,cuda=False,activation=activation,random_weights=True)
            d=a.derpar(x)
            dc=ANN(layers,cuda=True,activation=activation,random_weights=True).setpar(a.getpar()).derpar(x)
            self.assertAlmostEqual(d[0],dc[0],places=5)
            for i in range(len(d[1])):
                self.assertAlmostEqual(d[1][i],dc[1][i],places=5)

        def _test_layers_vec(self,layers,activation='softplus'):
            np.random.seed(1977)
            x=np.random.normal(size=(10,layers[0]))
            a=ANN(layers,cuda=False,activation=activation,random_weights=True)
            d=a.derpar(x)
            dc=ANN(layers,cuda=True,activation=activation,random_weights=True).setpar(a.getpar()).derpar(x)
            for i in range(len(d[0])):
                self.assertAlmostEqual(d[0][i],dc[0][i],places=5)
            for i in range(len(d[1])):
                for j in range(len(d[1][i])):
                    self.assertAlmostEqual(d[1][i,j],dc[1][i,j],places=5)

        def test_ann1(self):
            self._test_layers([10])
        def test_ann2(self):
            self._test_layers([10,10])
        def test_ann3(self):
            self._test_layers([10,10,10])
        def test_ann4(self):
            self._test_layers([10,8,6,4,2])

        def test_ann1r(self):
            self._test_layers([10],activation='relu')  # might fail due to relu discontinuity
        def test_ann2r(self):
            self._test_layers([10,10],activation='relu')  # might fail due to relu discontinuity
        def test_ann3r(self):
            self._test_layers([10,10,10],activation='relu')  # might fail due to relu discontinuity
        def test_ann4r(self):
            self._test_layers([10,8,6,4],activation='relu')  # might fail due to relu discontinuity

        def test_ann1_vec(self):
            self._test_layers_vec([10])
        def test_ann2_vec(self):
            self._test_layers_vec([10,10])
        def test_ann3_vec(self):
            self._test_layers_vec([10,10,10])
        def test_ann4_vec(self):
            self._test_layers_vec([10,8,6,4,2])

        def test_ann1r_vec(self):
            self._test_layers_vec([10],activation='relu')  # might fail due to relu discontinuity
        def test_ann2r_vec(self):
            self._test_layers_vec([10,10],activation='relu')  # might fail due to relu discontinuity
        def test_ann3r_vec(self):
            self._test_layers_vec([10,10,10],activation='relu')  # might fail due to relu discontinuity
        def test_ann4r_vec(self):
            self._test_layers_vec([10,8,6,4],activation='relu')  # might fail due to relu discontinuity

        def backprop(self,layers=None,activation="softplus"):
            if layers is None:
                layers=[10,8,6,4,2]
            np.random.seed(1977)
            ann=ANN(layers,activation=activation,random_weights=True,cuda=True)
            traj=np.random.normal(size=(1000,ann.narg))
            d=ann.derpar(traj)
            reference=np.matmul(d[0],d[1])
            state=ann.forward(traj)
            der=ann.backward_par(state.f,state)
            self.assertAlmostEqual(np.sum((reference-der)**2),0.0,places=5)

        def test_backprop1(self):
            self.backprop([10,10,10,10],"softplus")
        def test_backprop2(self):
            self.backprop([10,8,6,4,2],"softplus")
        def test_backprop1r(self):
            self.backprop([10,10,10,10],"relu")
        def test_backprop2r(self):
            self.backprop([10,8,6,4,2],"relu")


if __name__ == "__main__":
    unittest.main()



import unittest
import bussilab.cli as cli

class Test(unittest.TestCase):
    def test_check(self):
        self.assertIsNone(cli.cli("check"))
        self.assertIsNone(cli.cli("check --import"))
        self.assertIsNone(cli.cli(["check", "--import"]))
        self.assertIsNone(cli.cli("--version"))
        self.assertIsNone(cli.cli("list"))
        self.assertIsNone(cli.cli(""))

        self.assertEqual(cli.cli("not_existing_command",throws_on_parser_errors=False), 2)
        self.assertEqual(cli.cli(["check","not_existing_option"],throws_on_parser_errors=False), 2)

        with self.assertRaises(TypeError):
            cli.cli("not_existing_command")

        with self.assertRaises(TypeError):
            cli.cli(["check","not_existing_option"])

        with self.assertRaises(OSError):
            cli.cli("wham -b not_existing_file")

        self.assertEqual(cli.cli(["wham", "-b"], throws_on_parser_errors=False), 2)  # here argument is missing

        with self.assertRaises(TypeError):
            cli.cli(["wham", "-b"])  # here argument is missing

        with self.assertRaises(OSError):
            cli.cli(["wham", "-b", ""])  # "" is a file with no name

        with self.assertRaises(NotImplementedError):
            cli._list_submodules()

        self.assertIsNone(cli.cli("required"))
        self.assertIsNone(cli.cli("required --conda"))
        self.assertIsNone(cli.cli("required --macports --pyver 38"))

    def test_decorator(self):
        @cli.command("x")
        def _x():
            pass
        with self.assertRaises(NotImplementedError):
            _x()

if __name__ == "__main__":
    unittest.main()



"""
Module with some clustering tools
"""

from typing import Optional

import networkx
import numpy as np

from .coretools import Result
from .coretools import import_numba_jit

numba_jit=import_numba_jit()

class ClusteringResult(Result):
    """Result of a `bussilab.clustering` calculation."""
    def __init__(self,
                *,
                method: str,
                clusters: list,
                weights: Optional[list]
                ):
        self.method = method
        """`str` containing the name of the method used."""
        self.clusters = clusters
        """`list of lists` containing the members of each cluster."""
        self.weights = weights
        """`list` containing the weights of the clusters."""

def max_clique(adj,weights=None,*,min_size=0,max_clusters=None,use_networkit=False):
    """Clustering algorithm used in [Reisser et al, NAR (2020)](https://doi.org/10.1093/nar/gkz1184).

       Parameters
       ----------

       adj : array_like, square matrix

           adj[i,j] contains 1 (or True) if frames i and j are adjacent, 0 (or False) otherwise.

       weights : array_like, optional

           weights[i] contains the weight of the i-th frame.

       min_size : number

           Minimum cluster size. Clusters smaller than this size are not reported.
           When using weights, the cluster size is defined as the sum of the weights of
           the members of the cluster.

       max_clusters : int

           Maximum number of clusters.

       use_networkit : bool, optional

           if True, use a networkit implementation that seems to be faster.
           It requires python package networkit to be installed in advance!

       Example
       -------

       ```
       import scipy.spatial.distance as distance
       dist=distance.squareform(distance.pdist(trajectory))
       clustering.max_clique(dist<0.7)
       ```
    """
    # weights: optional weights
    # if adj is a graph, it will be copied
    cliques=[]
    ww=[]
    if use_networkit:
        import networkit # pylint: disable=import-error
        graph=networkit.nxadapter.nx2nk(networkx.Graph(adj))
        graph.removeSelfLoops()
        while graph.numberOfNodes()>0:
            if weights is None:
                cl=networkit.clique.MaximalCliques(graph,maximumOnly=True)
                cl.run()
                maxi=cl.getCliques()[0]
                maxw=len(maxi)
            else:
                cl=networkit.clique.MaximalCliques(graph)
                cl.run()
                maxw=0.0
                for i in cl.getCliques():
                    w=np.sum(weights[i])
                    if w > maxw:
                        maxi=i
                        maxw=w
            if maxw<min_size:
                break
            cliques.append(maxi)
            ww.append(maxw)
            if max_clusters is not None:
                if len(cliques)>=max_clusters:
                    break
            for i in maxi:
                graph.removeNode(i)
    else:
        graph=networkx.Graph(adj)
        while graph.number_of_nodes()>0:
            maxw=0.0
            for i in networkx.algorithms.clique.find_cliques(graph):
                if weights is not None:
                    w=np.sum(weights[i])
                else:
                    w=len(i)
                if w > maxw:
                    maxi=i
                    maxw=w
            if maxw<min_size:
                break
            cliques.append(maxi)
            ww.append(maxw)
            if max_clusters is not None:
                if len(cliques)>=max_clusters:
                    break
            graph.remove_nodes_from(maxi)
    return ClusteringResult(method="max_clique",clusters=cliques, weights=ww)

def daura(adj,weights=None,*,min_size=0,max_clusters=None):
    """Clustering algorithm introduced in Daura et al, Angew. Chemie (1999).

       WARNING: important fix in v0.0.39 for version with weights

       Parameters
       ----------

       adj : array_like, square matrix

           adj[i,j] contains 1 (or True) if frames i and j are adjacent, 0 (or False) otherwise.

       weights : array_like, optional

           weights[i] contains the weight of the i-th frame.

       min_size : number

           Minimum cluster size. Clusters smaller than this size are not reported.
           When using weights, the cluster size is defined as the sum of the weights of
           the members of the cluster.

       max_clusters : int

           Maximum number of clusters.

       Example
       -------

       ```
       import scipy.spatial.distance as distance
       dist=distance.squareform(distance.pdist(trajectory))
       clustering.daura(dist<0.7)
       ```
    """
    adj=adj.copy()  # take a copy (adj is modified while clustering)
    indexes=np.arange(len(adj))
    clusters=[]
    ww=[]
    if weights is not None:
        weights = weights.copy()  # take a copy (weights is modified while clustering)
    while len(indexes)>0:
        if weights is not None:
            d=np.sum(adj*weights[:,np.newaxis],axis=0)
        else:
            d=np.sum(adj,axis=0)
        n=np.argmax(d)
        if d[n]<min_size:
            break
        ww.append(d[n])
        ii=np.where(adj[n]>0)[0]
        clusters.append(indexes[ii])
        if max_clusters:
            if len(clusters) >= max_clusters:
                break
        adj=np.delete(adj,ii,axis=0)
        adj=np.delete(adj,ii,axis=1)
        if weights is not None:
            weights=np.delete(weights,ii)
        indexes=np.delete(indexes,ii)
    return ClusteringResult(method="daura",clusters=clusters, weights=ww)

@numba_jit
def _qt_outer(distances,next_,cutoff):
    candidates=np.empty(len(distances),dtype='int')
    n=0
    for i in range(next_):
        if distances[next_,i]<cutoff:
           candidates[n]=i
           n+=1
    for i in range(next_+1,len(distances)):
        if distances[next_,i]<cutoff:
           candidates[n]=i
           n+=1
    return candidates[:n]

@numba_jit
def _qt_inner(distances,dist_from_cluster,candidates,cutoff,weights):
    if len(dist_from_cluster)<1:
        return (-1,np.inf)
    next_i=0
    minval=dist_from_cluster[0]
    weight_minval=weights[candidates[0]]
    N=len(dist_from_cluster)
    for i in range(1,N):
        val=dist_from_cluster[i]
        if val<minval or (val==minval and weights[candidates[i]]>weight_minval):
            next_i=i
            minval=val
            weight_minval=weights[candidates[i]]
    if minval>cutoff:
        return (-1,minval)
    next_=candidates[next_i]
    for i in range(N):
        val=distances[next_][candidates[i]]
        if dist_from_cluster[i]<val:
            dist_from_cluster[i]=val
    dist_from_cluster[next_i]=np.inf
    return (next_,minval)

def qt(distances,cutoff,weights=None,*,min_size=0,max_clusters=None):
    """Quality threshold clustering.

       The method is explained in the [original paper](https://doi.org/10.1101/gr.9.11.1106).
       The implementation has been adapted from [this one](https://github.com/rglez/QT),
       which is also released under a GPL licence.
       Thus, if you use this algorithm please cite [this article](https://doi.org/10.1021/acs.jcim.9b00558),
       which also discusses the important differences between this algorithm and the Daura et al algorithm
       in the context of analysing molecular dynamics simulations.
       Additionally, mention which exact version of the bussilab package you used.

       The implementation included here, at variance with the original one, allows passing weights
       and can be used with arbitrary metrics. In addition, it also reports clusters of size 1
       (unless one passes max_clusters>1). The code is optimized when compared with the original one,
       and speed can be further increased by passing `np.array(distances,dtype='float32')` to distances.

       WARNING: important fix in v0.0.39 for version with weights

       As of version v0.0.40, clusters with the same number of members are prioritized based on their
       diameter (smaller diameter gets the priority). This is expected to make non-weighted calculations
       more reproducible, but might change some results when compared with previous versions.
       In addition, when growing a single candidate cluster, it two points are at the same
       distance from the growing cluster the one with higher weight is chosen. With these
       priorities, any choice where either (a) weights are float or (b) distances are float
       should lead to deterministing clustering irrespective of roundoff errors, assuming
       floats are never identical.


       Parameters
       ----------

       distances : array_like, square matrix

           distances[i,j] contains the distance between i and j frame.

       cutoff : number

           maximum distance for two frames to be included in the same cluster

       weights : array_like, optional

           weights[i] contains the weight of the i-th frame.

       min_size : number

           Minimum cluster size. Clusters smaller than this size are not reported.
           When using weights, the cluster size is defined as the sum of the weights of
           the members of the cluster.

       max_clusters : int

           Maximum number of clusters.

       Example
       -------

       ```
       import scipy.spatial.distance as distance
       dist=distance.squareform(distance.pdist(trajectory))
       clustering.qt(dist,0.7)

       clustering.qt(np.array(dist,dtype='float32')) # should be slightly faster
       ```
    """
    clusters=[]
    ww=[]
    N=len(distances)
    if weights is None:
        weights=np.ones(N,dtype="int")
    else:
        weights=weights.copy()
    distances=distances.copy()
    np.fill_diagonal(distances,0.0)
    indexes=np.arange(N)

    ncluster=0
    while len(weights)>0:
        degrees=np.sum((distances < cutoff)*weights[:,np.newaxis],axis=0)
        sorted_indexes=np.argsort(-degrees)

        cluster_size=0
        diameter=0.0
        cluster=[]
        i_degrees=0
        for i_degrees in range(len(sorted_indexes)):

            if degrees[sorted_indexes[i_degrees]] < cluster_size: # optimization
                break

            next_=sorted_indexes[i_degrees]
            precluster = [next_]
            new_cluster_diameter=0.0
            candidates=_qt_outer(distances,next_,cutoff)
            dist_from_cluster=distances[next_][candidates]
            while True:
                (next_,minval)=_qt_inner(distances,dist_from_cluster,candidates,cutoff,weights)
                if(next_<0):
                    break
                precluster.append(next_)
                if minval > new_cluster_diameter:
                    new_cluster_diameter = minval
            new_cluster_size=np.sum(weights[precluster])
            # pick largest cluster (sum of weights)
            # if same size, pick the most compact one (smaller diameter)
            if new_cluster_size > cluster_size or (new_cluster_size == cluster_size and new_cluster_diameter < diameter):
                cluster_size = new_cluster_size
                cluster = precluster
                diameter = new_cluster_diameter

        if cluster_size < min_size:
            break

        ncluster += 1

        clusters.append(indexes[cluster])
        ww.append(cluster_size)

        if max_clusters is not None:
            if ncluster >= max_clusters:
                break

        # see https://stackoverflow.com/questions/63563151/how-to-remove-columns-and-rows-at-certain-indexes-in-numpy-at-the-same-time
        idx = list(set(range(len(distances))).difference(cluster))
        distances=distances[np.ix_(idx,idx)]
        weights=weights[idx]
        indexes=indexes[idx]
    return ClusteringResult(method="qt",clusters=clusters, weights=ww)



"""
Module implementing Slack notifications.

This module sends notification through an App installed in the Slack workspace.
Some settings are needed first for authentication. It is recommended to add a
file named `.bussilabrc` to your home directory with the following content:
```bash
notify:
  token: xoxb-00000
  channel: U00000
```
The `token` here should be provided by the administrator of your workspace.
The channel should be the Slack ID associated to your user. It can be found
looking in your Slack profile. With these settings, the tool will send
notifications to you by default.

Notifications can then be sent using either the command line:
```bash
bussilab notify --message "text here"
```
or from python:
```python
from bussilab.notify import notify
notify("text here")
```

Notice that the message is optional. Even with an empty message, the footer
will allow you to reconstruct from which machine and directory the message was
sent from. This might be sufficient for your goal.

You can also indicate a specific channel for the notification using the
`channel` option:
```bash
bussilab notify --message "text here" --channel "project-myproject"
```
or from python:
```python
from bussilab.notify import notify
notify("text here", channel="project-myproject")
```
This will only work if the App has been added to the specified channel.

The following syntax can be used to upload a file:
```bash
bussilab notify --message "text here" --file /path/to/file
```
or from python:
```python
from bussilab.notify import notify
notify("text here",file="/path/to/file")
```

The commands above will return the URL of the message. This URL can be used
later to update or delete them or to post reactions:
```bash
url=$(bussilab notify --message "text here")
bussilab notify --update $url --message "revised message"
bussilab notify --react $url:heart

# this will remove only the reaction:
bussilab notify --delete $url:heart

# this will remove the entire message:
bussilab notify --delete $url

url=$(bussilab notify --message "text here")
```
or from python:
```python
from bussilab.notify import notify
url=notify("text here")
notify("revised message", update=url)
notify(react=url+":heart")
notify(delete=url+":heart")
notify(delete=url)
```
In these cases, the channel is not needed and should not be provided.
Notice that you will only be able to update or delete messages sent through the
App.
"""

import datetime
import re
import os
import socket
import time
import warnings

try:
    # slack client 3
    from slack_sdk import WebClient
    from slack_sdk.web.base_client import SlackResponse
    from slack_sdk.errors import SlackApiError
except ModuleNotFoundError:
    # slack client 2
    from slack import WebClient
    from slack.web.base_client import SlackResponse
    from slack.errors import SlackApiError

from . import coretools

from typing import cast

def _try_multiple_times(func,*args,**kwargs):
     max_attempts=10
     num_attempts=0
     while True:
        try:
            num_attempts+=1
            return func(*args,**kwargs)
        except SlackApiError as e:
            if num_attempts>=max_attempts:
                raise

            if "error" in e.response and e.response["error"]=="ratelimited" and "Retry-After" in e.response.headers:
                warnings.warn("Slack API, retry-after "
                              +str(e.response.headers["Retry-After"])
                              +" seconds"+
                              " ["+str(num_attempts)+"/"+str(max_attempts)+"]",
                              UserWarning)
                time.sleep(float(e.response.headers["Retry-After"]))
            elif not hasattr(e.response,"status_code") or e.response.status_code!=200:
                warnings.warn("Slack API, server-side problem: "
                              +str(e.response)+"\n"+
                              "retrying after 30 seconds"+
                              " ["+str(num_attempts)+"/"+str(max_attempts)+"]",
                              UserWarning)
                time.sleep(30)
            else:
                raise

_match_message=r"https://[^/]*\.slack\.com/archives/.*"

def _parse_url(url: str):
    if re.match(r"^https://[^/]*\.slack\.com/archives/.*:.*",url):
        organization = re.sub(r"^https://","", re.sub(r"\.slack\.com/archives/.*","",url))
        url=re.sub(r"^https://[^/]*\.slack\.com/archives/","",url)
        url1=re.sub(r":.*","",url)
        url1=url1[:-6]+"."+url1[-6:]
        channel=re.sub("/p.*","",url1)
        ts=re.sub(":.*$","",re.sub("^.*/p","",url1))
        react=re.sub(r".*:","",url)
        return { "type":"reaction", "ts":ts, "channel":channel, "organization":organization, "reaction": react}
    if re.match(r"^https://[^/]*\.slack\.com/archives/.*",url):
        organization = re.sub(r"^https://","", re.sub(r"\.slack\.com/archives/.*","",url))
        url=re.sub(r"^https://[^/]*\.slack\.com/archives/","",url)
        url=re.sub(r"\?.*","",url)
        url=url[:-6]+"."+url[-6:]
        channel=re.sub("/p.*","",url)
        ts=re.sub(".*/p","",url)
        return { "type":"message", "ts":ts, "channel":channel, "organization":organization }
    if re.match(r"^https://[^/]*\.slack\.com/files/.*",url):
        organization = re.sub(r"^https://","", re.sub(r"\.slack\.com/files/.*","",url))
        url=re.sub(r"^https://[^/]*\.slack\.com/files/","",url)
        user=re.sub("/.*","",url)
        id=re.sub("^"+user+"/","",url)
        id=re.sub("/.*","",id)
        return { "type":"file", "id":id, "user":user, "organization":organization }
    return {}

def notify(message: str = "",
           channel: str = None,
           *,
           react: str = None,
           update: str = None,
           delete: str = None,
           reply: str = None,
           reply_broadcast: str = None,
           title: str = "",
           screenlog: str = "",
           screenlog_maxlines: int = 0,
           footer: bool = True,
           type: str = "mrkdwn",
           file: str = "",
           token: str = None):
    """Tool to send notifications to Slack.

       Parameters
       ----------

       message: str

           A string that will form the body of the message.

       channel: None or str

           The channel. By default, taken from your `~/.bussilabrc`
           configuration file.

       update: None or str

           The URL of a message to be updated.

       delete: None or str

           The URL of a message to be deleted. By passing a URL
           concatenated with the string `":name_of_reaction"` you can
           delete a reaction. Buy passing two comma-separated URLs
           you can delete both a file and the message with which it was
           shared.
       
       reply: None or str
       
           The URL of a message to be replied
       
       reply_broadcast: None or str
       
           The URL of a message to be broadcast-replied
       
       react: None or str
       
           The URL of a message to which you want to add a reaction,
           followed by the string `:name_of_the_reaction`
           
       file: None or str
       
           The path of a file to be uploaded

       title: str

           The title of the notification.

       footer: bool

           If True, a footer is added with current user, machine, and
           directory.

       type: str

           The type of message. Can be "mrkdwn" or "plain_text".

       token: None or str

           The token. By default, taken from your `~/.bussilabrc`
           configuration file.

       Returns
       -------

           str
               A string with the URL of the sent message.
               In case the `delete` keyword is used, it returns an empty
               string.
               In case a file is uploaded, it returns two comma-separated
               URLs corresponding to the message and to the file.


       Example
       -------

       ```python
       from bussilab.notify import notify
       notify("send this message")
       ```
       See `bussilab.notify` for more examples.
    """

    if [bool(channel),
        bool(update),
        bool(react),
        bool(delete),
        bool(reply),
        bool(reply_broadcast)
       ].count(True)>1:
        raise TypeError("channel/update/delete/reply/reply_broadcast are mutually incompatible")

    if len(file)>0 and (update or react or delete or reply_broadcast):
        raise TypeError("files cannot be updated")

    config = None
    if token is None:
        config = coretools.config()
        token=config["notify"]["token"]

    client = WebClient(token=token)

    if delete:
        # this is to enable deletion of both a message and a file:
        delete_multi=delete.split(",")
        if len(delete_multi)>1:
            for d in delete_multi:
                notify(message,channel,delete=d,token=token)
            return ""
        delete_dict=_parse_url(delete)
        if not delete_dict:
            raise TypeError("cannot parse delete URL")
        if delete_dict["type"]=="message":
            _try_multiple_times(client.chat_delete,
                                channel=delete_dict["channel"],
                                ts=delete_dict["ts"])
        elif delete_dict["type"]=="file":
            _try_multiple_times(client.files_delete,
                                file=delete_dict["id"])
        elif delete_dict["type"]=="reaction":
            _try_multiple_times(client.reactions_remove,
                                channel=delete_dict["channel"],
                                timestamp=delete_dict["ts"],
                                name=delete_dict["reaction"])
        else:
            raise RuntimeError("unknown type")
        # delete always returns an empty string
        return ""

    if react:
        react_dict=_parse_url(react)
        response = _try_multiple_times(client.reactions_add,
          name=react_dict["reaction"],
          timestamp=react_dict["ts"],
          channel=react_dict["channel"])
        return react

    screenlog_message=""
    if len(screenlog)>0:
        # we manually removed "deleted" lines.
        # this is very useful for tdqm-like logs
        with open(screenlog,'rb') as handler:
            screenlog_message=handler.read().decode()
       	    screenlog_message=re.sub(r'.*\r([^\n])', r'\1', screenlog_message, flags=re.M)
        if screenlog_maxlines>0:
            screenlog_message_lines=screenlog_message.split("\n")
            if len(screenlog_message_lines) > screenlog_maxlines:
                screenlog_message_lines = screenlog_message_lines[-screenlog_maxlines:]
            screenlog_message="\n".join(screenlog_message_lines)

    if len(screenlog_message)>2900:
        screenlog_message=screenlog_message[:2900] + " [truncated]"
        
    if len(message)>2900:
        message=message[:2900] + " [truncated]"

    if len(title)>2900:
        title=title[:2900] + " [truncated]"

    if update:
        update_dict=_parse_url(update)
        if not update_dict:
           raise TypeError("")
        organization=update_dict["organization"]
    elif reply:
        reply_dict=_parse_url(reply)
        organization=reply_dict["organization"]
    elif reply_broadcast:
        reply_dict=_parse_url(reply_broadcast)
        organization=reply_dict["organization"]
    else:
        if channel is None:
            if config is None:
                config = coretools.config()
            channel=config["notify"]["channel"]
        if re.match(r"^https://[^/]*\.slack\.com/archives/.*", channel):
            organization = re.sub("^https://","", re.sub(r"\.slack\.com/archives/.*","",channel))
            channel=re.sub(r"^https://[^/]*\.slack\.com/archives/","",channel)
        else:
            # this is needed to set organization correctly (so as to build the
            # proper link) when passing the name of a channel
            organization = ""

    blocks=[]
    text=""

    if len(title) > 0:
        text+="*" + title+"*\n"
        blocks.append(
           {
               "type": "section",
               "text": {"type": "mrkdwn", "text": "*" + title + "*"},
           }
           )

    if len(message) > 0:
        text+=message+"\n"
        blocks.append(
           {
               "type": "section",
               "text": {
                         "type": type,
                         "text": message
                       },
           }
           )
        
    if len(screenlog_message) > 0:
        text+=screenlog_message+"\n"
        blocks.append(
           {
               "type": "section",
               "text": {
                         "type": "mrkdwn",
                         "text": "```\n" + screenlog_message + "\n```\n"
                       },
           }
           )

    if footer:
        footer_text = ""
        if update:
            footer_text += "Updated"
        else:
            footer_text += "Sent"
        footer_text += " by "+ os.environ['USER']
        footer_text += " at " + socket.gethostname() +'\n'
        footer_text += "pwd: " + os.getcwd() + '\n'
        footer_text += datetime.datetime.now().isoformat(sep=' ',timespec='milliseconds')
        text+=footer_text+"\n"
        blocks.append({
                          "type": "context",
                          "elements": [
                              {  # type: ignore
                                "type": "mrkdwn",
                                "text": footer_text
                              }
                          ]
                      })
    if len(blocks)==0:
        text+="(empty notification)"
        blocks.append({
                          "type": "section",
                          "text": {
                                     "type": type,
                                     "text": "(empty notification)"
                                  }
                      })

    if update:
        response = _try_multiple_times(client.chat_update,
                   channel=update_dict["channel"],
                   text=text,
                   blocks=blocks,
                   ts=update_dict["ts"])
    elif len(file)>0:
        initial_comment = ""
        if len(title)>0:
            initial_comment += "*" + title + "*\n"
        if len(message)>0:
            initial_comment += message +"\n"
        if footer:
            initial_comment += footer_text

        v2=False
# # Note:
# # UserWarning: client.files_upload() may cause some issues like timeouts for
# # relatively large files. Our latest recommendation is to use client.files_upload_v2(),
# # which is mostly compatible and much stabler, instead.
# # However, to me v2 does not work (the file is not accessible and not posted to the channel)
#         try:
#             x=client.files_upload_v2
#             v2=True
#         except AttributeError:
#             pass


        if v2:
            response = _try_multiple_times(client.files_upload_v2,file=file,title=file,channel=channel)
        else:
            if reply:
                response = _try_multiple_times(client.files_upload,
                                               file=file,
                                               channels=reply_dict["channel"],
                                               title=file,
                                               thread_ts=reply_dict["ts"])
            else:
                response = _try_multiple_times(client.files_upload,
                                               file=file,
                                               title=file,
                                               channels=channel,
                                               initial_comment=initial_comment)
    elif reply:
        response = _try_multiple_times(client.chat_postMessage,
                   blocks=blocks,
                   text=text,
                   channel=reply_dict["channel"],
                   thread_ts=reply_dict["ts"])
    elif reply_broadcast:
        response = _try_multiple_times(client.chat_postMessage,
                   blocks=blocks,
                   text=text,
                   channel=reply_dict["channel"],
                   thread_ts=reply_dict["ts"],
                   reply_broadcast=True)
    else:
        response = _try_multiple_times(client.chat_postMessage,
                   blocks=blocks,
                   text=text,
                   channel=channel)

    response = cast(SlackResponse, response)

    if len(organization)==0:
        base_url=_try_multiple_times(client.auth_test)["url"]
    else:
        base_url="https://" + organization + ".slack.com/"

    if len(file)==0:
        url=base_url + "archives/" + response["channel"] + "/p" + response["ts"][:-7] + response["ts"][-6:]
    else:
        
        k=list(response["file"]["shares"].keys())[0] # empirically, pick the first one. There should be only one!
        channel=list(response["file"]["shares"][k].keys())[0] # empirically, pick the first one. There should be only one!
        ts=response["file"]["shares"][k][channel][0]["ts"]
        url=base_url + "archives/" + channel + "/p" + ts[:-7] + ts[-6:]
        url+="," + base_url + "files/" + response["file"]["user"] + "/" + response["file"]["id"]

    return url
